#include <stdexcept>

#include <boost/tokenizer.hpp>

#include <iostream>
#include "easy_curl.h"

#define DEBUG_CURL

#ifdef DEBUG_CURL

#include <sstream>
#include <boost/thread/mutex.hpp>

namespace
{
	boost::mutex date_mutex;

	std::string get_name(CURL* handle)
	{
		char buf[30];
		sprintf(buf, "%u.log", handle);
		return buf;
	}

	std::string get_time()
	{
		boost::mutex::scoped_lock lock(date_mutex);

		// Get date
		char buffer[256];
		time_t timestamp;
		time(&timestamp);
		tm* now = localtime(&timestamp);
		sprintf(buffer, "%02d:%02d:%02d", now->tm_hour, now->tm_min, now->tm_sec);

		return buffer;
	}

	void call_debug(CURL* handle, const std::string& text)
	{
		std::ofstream out(get_name(handle).c_str(), std::ios::app);
		out << std::endl << get_time() << "| -------- [CALL] " << text << " --------" << std::endl << std::endl;
		out.close();
	}

	int curl_debug(CURL* handle, curl_infotype type, char* str, size_t len, void *)
	{
		std::ofstream out(get_name(handle).c_str(), std::ios::app);

		out << get_time() << "| ";

		switch (type)
		{
			case CURLINFO_TEXT:
			{
				out << "[info] ";
				break;
			}

			case CURLINFO_HEADER_OUT:
			{
				out << ">> send: ";
				break;
			}

			case CURLINFO_DATA_OUT:
			{
				out << ">> send data (skip log)" << std::endl;
				return 0;
				break;
			}

			case CURLINFO_HEADER_IN:
			{
				out << "<< recv: ";
				break;
			}

			case CURLINFO_DATA_IN:
			{
				out << "<< recv data (skip log)" << std::endl;
				return 0;
				break;
			}
		}

		out << std::string(str, len);
		out.close();
		
		return 0;
	}
}
#endif

easy_curl::easy_curl() : m_curl(curl_easy_init()), m_fresh_connect(false)
{
	reset();
}

easy_curl::~easy_curl()
{
	curl_easy_cleanup(m_curl);
}

void easy_curl::set_fresh_connect(bool fresh)
{
	m_fresh_connect = fresh;
}

void easy_curl::http_get_per_line(const std::string& url, const boost::function<void(const std::string& line)>& reader)
{
	m_line_reader = reader;
	http_get(url);
	m_line_reader.clear();
}

const std::string easy_curl::http_get(const std::string& url)
{
	#ifdef DEBUG_CURL
		call_debug(m_curl, "http_get(" + url + ")");
	#endif

	reset();

	if(!m_curl)
	{
		m_error = CURLE_FAILED_INIT;
		return "";
	}

	curl_easy_setopt(m_curl, CURLOPT_URL, url.c_str());
	curl_easy_setopt(m_curl, CURLOPT_HEADER, 0);
	curl_easy_setopt(m_curl, CURLOPT_WRITEDATA, this);
	curl_easy_setopt(m_curl, CURLOPT_WRITEFUNCTION, easy_curl::contents_builder);

	m_error = curl_easy_perform(m_curl);

	curl_easy_getinfo(m_curl, CURLINFO_RESPONSE_CODE, &m_http_code);

	return m_contents;
}

const std::string easy_curl::ftp_list(const std::string& url)
{
	#ifdef DEBUG_CURL
		call_debug(m_curl, "ftp_list(" + url + ")");
	#endif

	reset();

	if(!m_curl)
	{
		m_error = CURLE_FAILED_INIT;
		return "";
	}

	curl_easy_setopt(m_curl, CURLOPT_URL, url.c_str());
	curl_easy_setopt(m_curl, CURLOPT_HEADER, 0);
	curl_easy_setopt(m_curl, CURLOPT_WRITEDATA, this);
	curl_easy_setopt(m_curl, CURLOPT_WRITEFUNCTION, easy_curl::list_builder);
	curl_easy_setopt(m_curl, CURLOPT_FTP_CREATE_MISSING_DIRS, 1);

	m_error = curl_easy_perform(m_curl);

	return m_list;
}

CURLcode easy_curl::download(const std::string& from, const std::string& to)
{
	#ifdef DEBUG_CURL
		call_debug(m_curl, "download(" + from + ", " + to + ")");
	#endif

	reset();

	if(!m_curl) 
		return m_error = CURLE_FAILED_INIT;

	m_outfile.clear();
	m_outfile.open(to.c_str(), std::ios::binary);
	if(!m_outfile.is_open())
		throw std::runtime_error("Cannot open \"" + to + "\" for writing.");
	
	curl_easy_setopt(m_curl, CURLOPT_URL, from.c_str());
	curl_easy_setopt(m_curl, CURLOPT_HEADER, 0);
	curl_easy_setopt(m_curl, CURLOPT_WRITEDATA, this);
	curl_easy_setopt(m_curl, CURLOPT_WRITEFUNCTION, easy_curl::data_writer);

	m_error = curl_easy_perform(m_curl);
	m_outfile.close();

	if(m_error == CURLE_OK)
	{
		curl_easy_getinfo(m_curl, CURLINFO_SPEED_DOWNLOAD, &m_speed);
		curl_easy_getinfo(m_curl, CURLINFO_TOTAL_TIME, &m_time);
	}
	return m_error;
}

CURLcode easy_curl::upload(const std::string& from, const std::string& to)
{
	#ifdef DEBUG_CURL
		call_debug(m_curl, "upload(" + from + ", " + to + ")");
	#endif

	reset();

	if(!m_curl) 
		return m_error = CURLE_FAILED_INIT;

	m_infile.clear();
	m_infile.open(from.c_str(), std::ios::binary);
	if(!m_infile.is_open())
		throw std::runtime_error("Cannot open \"" + to + "\" for reading.");

	curl_easy_setopt(m_curl, CURLOPT_URL, to.c_str());
	curl_easy_setopt(m_curl, CURLOPT_HEADER, 0);
	curl_easy_setopt(m_curl, CURLOPT_UPLOAD, 1);
	curl_easy_setopt(m_curl, CURLOPT_FTP_CREATE_MISSING_DIRS, 1);
	curl_easy_setopt(m_curl, CURLOPT_READDATA, this);
	curl_easy_setopt(m_curl, CURLOPT_READFUNCTION, easy_curl::data_reader);

	m_error = curl_easy_perform(m_curl);
	m_infile.close();
	
	if(m_error == CURLE_OK)
	{
		curl_easy_getinfo(m_curl, CURLINFO_SPEED_UPLOAD, &m_speed);
		curl_easy_getinfo(m_curl, CURLINFO_TOTAL_TIME, &m_time);
	}
	return m_error;
}

CURLcode easy_curl::ftp_cmd(const std::string& url, const std::string& cmd)
{
	#ifdef DEBUG_CURL
		call_debug(m_curl, "ftp_cmd(" + url + ", " + cmd + ")");
	#endif

	reset();

	if(!m_curl) 
		return m_error = CURLE_FAILED_INIT;

	curl_easy_setopt(m_curl, CURLOPT_NOBODY, 1);
	curl_easy_setopt(m_curl, CURLOPT_HEADER, 0);
	curl_easy_setopt(m_curl, CURLOPT_URL, url.c_str());
	
	curl_slist* list = NULL;
	list = curl_slist_append(list, cmd.c_str());
	curl_easy_setopt(m_curl, CURLOPT_POSTQUOTE, list);
	
	m_error = curl_easy_perform(m_curl);

	curl_slist_free_all(list);
  
	return m_error;
}

void easy_curl::reset()
{
	m_http_code = 0;
	m_speed = 0.0;
	m_time  = 0.0;
	m_list.clear();
	m_contents.clear();
	m_error = CURLE_OK;

	curl_easy_reset(m_curl);
	if(m_fresh_connect)
	{
		#ifdef DEBUG_CURL
			call_debug(m_curl, "FRESH CONNECT");
		#endif
		curl_easy_setopt(m_curl, CURLOPT_FRESH_CONNECT, 1);
		//curl_easy_setopt(m_curl, CURLOPT_FORBID_REUSE, 1);
	}
	curl_easy_setopt(m_curl, CURLOPT_NOSIGNAL, 1);
	curl_easy_setopt(m_curl, CURLOPT_NOPROGRESS, 1);
	//curl_easy_setopt(m_curl, CURLOPT_MAXREDIRS, -1);

	#ifdef DEBUG_CURL
		curl_easy_setopt(m_curl, CURLOPT_VERBOSE, 1);
		curl_easy_setopt(m_curl, CURLOPT_DEBUGFUNCTION, curl_debug);
	#endif
}


CURLcode easy_curl::get_err_no() const
{
	return m_error;
}

std::string easy_curl::get_err_str() const
{
	return curl_easy_strerror(m_error);
}

double easy_curl::get_speed() const
{
	return m_speed;
}

double easy_curl::get_time() const
{
	return m_time;
}

long easy_curl::get_http_code() const
{
	return m_http_code;
}

std::string easy_curl::format(const std::string& str)
{ 
	std::string correct_str = str;
	std::string::size_type pos = 0;

	while((pos = correct_str.find("%", pos)) != std::string::npos)
	{
		correct_str.replace(pos, 1, "%25");
		pos++;
	}

	return correct_str;
}

size_t easy_curl::contents_builder(void* buffer, size_t size, size_t bytes, void* object)
{
	easy_curl* obj = reinterpret_cast<easy_curl*>(object);
	std::streamsize count = size * bytes;
	char *p = reinterpret_cast<char*>(buffer);
	obj->m_contents.append(p, count);
	if(obj->m_line_reader)
	{
		std::string::size_type pos = obj->m_contents.find_last_of('\n');
		if(pos != std::string::npos)
		{
			typedef boost::char_separator<char> char_separator;
			typedef boost::tokenizer<char_separator> tokenizer;
			char_separator separators("\n");
			tokenizer tokens(obj->m_contents.substr(0, pos), separators);

			for(tokenizer::iterator it = tokens.begin(); it != tokens.end(); ++it)
				obj->m_line_reader(*it);

			obj->m_contents.erase(0, pos + 1);
		}
	}
	return count;
}

size_t easy_curl::data_writer(void* buffer, size_t size, size_t bytes, void* object)
{
	easy_curl* obj = reinterpret_cast<easy_curl*>(object);
	size_t count = size * bytes;
	obj->m_outfile.write(reinterpret_cast<const char*>(buffer), count);
	return (obj->m_outfile.good()) ? count : 0;
}

size_t easy_curl::data_reader(void* buffer, size_t size, size_t bytes, void* object)
{
	easy_curl* obj = reinterpret_cast<easy_curl*>(object);
	size_t count = size * bytes;
	obj->m_infile.read(reinterpret_cast<char*>(buffer), count);
	return obj->m_infile.gcount();
}

size_t easy_curl::list_builder(void* buffer, size_t size, size_t bytes, void* object)
{
	easy_curl* obj = reinterpret_cast<easy_curl*>(object);
	size_t count = size * bytes;
	const char* buf = reinterpret_cast<const char*>(buffer);
	obj->m_list.append(buf, count);
	return count;
}
